package state

import (
	"bytes"
	"context"
	"database/sql"
	"testing"
	"time"

	"github.com/filecoin-project/dealbot/tasks"
	"github.com/robfig/cron/v3"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestScheduledTask(t *testing.T) {
	const scheduledTasks = `[{"Miner":"t01000","PayloadCID":"bafk2bzacedli6qxp43sf54feczjd26jgeyfxv4ucwylujd3xo5s6cohcqbg36","CARExport":false,"Schedule":"*/5 * * * * *","ScheduleLimit":"30s"}]`

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	key, err := makeKey()
	require.NoError(t, err)

	runNotice := make(chan string, 1)

	err = WipeAndReset(dbConn, migrator)
	require.NoError(t, err)
	stateInterface, err := newStateDBWithNotify(ctx, dbConn, migrator, "", key, nil, runNotice)
	require.NoError(t, err)
	state := stateInterface.(*stateDB)

	// Replace normal scheduler with one that accepts seconds
	state.cronSched.Stop()
	state.cronSched = cron.New(cron.WithSeconds())
	state.cronSched.Start()

	err = populateTestTasks(ctx, bytes.NewReader([]byte(scheduledTasks)), stateInterface)
	require.NoError(t, err)

	worker := "test_worker"
	task, scheduled, err := state.popTask(ctx, worker, tasks.InProgress, nil)
	require.NoError(t, err)
	require.Nil(t, task, "should not find unassigned task")

	t.Log("waiting for scheduled task 1st generation")
	var newTaskID string
	select {
	case <-time.After(10 * time.Second):
		t.Fatal("scheduled task did not run")
	case newTaskID = <-runNotice:
		t.Log("scheduler generated task", newTaskID)
	}

	newTask, err := state.Get(ctx, newTaskID)
	require.NoError(t, err)
	require.NotNil(t, newTask, "Did not find new generated task")
	assert.Equal(t, newTaskID, newTask.UUID.String(), "wrong uuid for new task")

	sch, _ := newTask.Schedule()
	assert.Equal(t, "", sch, "new task should not have schedule")

	t.Log("popping generated task")
	task, scheduled, err = state.popTask(ctx, worker, tasks.InProgress, nil)
	require.NoError(t, err)
	require.NotNil(t, task, "Did not find runable task")
	require.False(t, scheduled, "should not have found scheduled task")
	require.Equal(t, worker, task.WorkedBy.Must().String(), "should be assigned to test_worker")
	runCount, err := task.RunCount.AsInt()
	require.NoError(t, err)
	assert.Equal(t, 0, int(runCount))
	t.Log("popped task assigned to", worker)

	// Make task as in progress
	err = state.transact(ctx, func(tx *sql.Tx) error {
		_, err := tx.ExecContext(ctx, setTaskStatusSQL, newTaskID, tasks.InProgress.Int(), time.Now())
		return err
	})
	if err != nil {
		t.Fatal(err)
	}
	time.Sleep(time.Second)

	// Make task as complete
	err = state.transact(ctx, func(tx *sql.Tx) error {
		_, err := tx.ExecContext(ctx, setTaskStatusSQL, newTaskID, tasks.Failed.Int(), time.Now())
		return err
	})
	if err != nil {
		t.Fatal(err)
	}

	t.Log("waiting for scheduled task 2nd generation")
	select {
	case <-time.After(10 * time.Second):
		t.Fatal("scheduled task did not run again")
	case newTaskID = <-runNotice:
		t.Log("scheduler generated task", newTaskID)
	}
	newTask, err = state.Get(ctx, newTaskID)
	require.NoError(t, err)
	require.NotNil(t, newTask, "Did not find new generated task")
	assert.Equal(t, newTaskID, newTask.UUID.String(), "wrong uuid for new task")

	t.Log("popping next generated task")
	task, scheduled, err = state.popTask(ctx, worker, tasks.InProgress, nil)
	require.NoError(t, err)
	require.NotNil(t, task, "Did not find runable task")
	require.False(t, scheduled, "should not have found scheduled task")
	require.Equal(t, worker, task.WorkedBy.Must().String(), "should be assigned to test_worker")
	runCount, err = task.RunCount.AsInt()
	require.NoError(t, err)
	assert.Equal(t, 1, int(runCount))
	t.Log("popped another task assigned to", worker)

	// Make task as in progress
	err = state.transact(ctx, func(tx *sql.Tx) error {
		_, err := tx.ExecContext(ctx, setTaskStatusSQL, newTaskID, tasks.InProgress.Int(), time.Now())
		return err
	})
	if err != nil {
		t.Fatal(err)
	}

	t.Log("waiting to see that task is not scheduled for 3nd generation, because previous task incomplete")
	select {
	case <-time.After(6 * time.Second):
	case newTaskID = <-runNotice:
		t.Error("scheduler should not have generated another task")
	}

	// Make task as complete
	err = state.transact(ctx, func(tx *sql.Tx) error {
		_, err := tx.ExecContext(ctx, setTaskStatusSQL, newTaskID, tasks.Successful.Int(), time.Now())
		return err
	})
	if err != nil {
		t.Fatal(err)
	}

	t.Log("waiting to see task scheduled for 3nd generation, because previous task is complete")
	select {
	case <-time.After(6 * time.Second):
		t.Error("new task should have been generated by now")
	case newTaskID = <-runNotice:
	}
}

func TestScheduledTaskLimit(t *testing.T) {
	const scheduledTasks = `[{"Miner":"t01000","PayloadCID":"bafk2bzacedli6qxp43sf54feczjd26jgeyfxv4ucwylujd3xo5s6cohcqbg36","CARExport":false,"Schedule":"*/5 * * * * *","ScheduleLimit":"7s"}]`

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	key, err := makeKey()
	require.NoError(t, err)

	runNotice := make(chan string, 1)

	err = WipeAndReset(dbConn, migrator)
	require.NoError(t, err)
	stateInterface, err := newStateDBWithNotify(ctx, dbConn, migrator, "", key, nil, runNotice)
	require.NoError(t, err)
	state := stateInterface.(*stateDB)

	// Replace normal scheduler with one that accepts seconds
	state.cronSched.Stop()
	state.cronSched = cron.New(cron.WithSeconds())
	state.cronSched.Start()

	err = populateTestTasks(ctx, bytes.NewReader([]byte(scheduledTasks)), stateInterface)
	require.NoError(t, err)

	// Wait for any scheduling
	var newTaskID string
	select {
	case <-time.After(6 * time.Second):
		t.Error("scheduler did not generate new task")
	case newTaskID = <-runNotice:
		t.Log("got 1st generated task")
	}

	// Make task as complete so it does not block more generation
	err = state.transact(ctx, func(tx *sql.Tx) error {
		_, err := tx.ExecContext(ctx, setTaskStatusSQL, newTaskID, tasks.Failed.Int(), time.Now())
		return err
	})
	if err != nil {
		t.Fatal(err)
	}

	// Make sure there is no more scheduling (schedule expired)
	t.Log("waiting for more scheduling after schedule expiration")
	select {
	case <-time.After(6 * time.Second):
	case <-runNotice:
		t.Error("scheduler should not have generated another task")
	}

	count, err := state.recoverScheduledTasks(ctx)
	if err != nil {
		t.Fatal(err)
	}
	if count != 0 {
		t.Fatal("should not have recovered any scheduled tasks")
	}
}
